from fontTools.misc.testTools import parseXML
from fontTools.misc.textTools import deHexStr
from fontTools.misc.xmlWriter import XMLWriter
from fontTools.misc.fixedTools import floatToFixed as fl2fi
from fontTools.pens.statisticsPen import StatisticsPen
from fontTools.ttLib import TTFont, TTLibError
import fontTools.ttLib.tables.otTables as otTables
from fontTools.ttLib.tables._a_v_a_r import table__a_v_a_r
from fontTools.ttLib.tables._f_v_a_r import table__f_v_a_r, Axis
import fontTools.varLib.models as models
import fontTools.varLib.varStore as varStore
from io import BytesIO
import os
import unittest


DATA_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), "data")

TEST_DATA = deHexStr(
    "00 01 00 00 00 00 00 02 "
    "00 04 C0 00 C0 00 00 00 00 00 13 33 33 33 40 00 40 00 "
    "00 03 C0 00 C0 00 00 00 00 00 40 00 40 00"
)


class AxisVariationTableTest(unittest.TestCase):
    def assertAvarAlmostEqual(self, segments1, segments2):
        self.assertSetEqual(set(segments1.keys()), set(segments2.keys()))
        for axisTag, mapping1 in segments1.items():
            mapping2 = segments2[axisTag]
            self.assertEqual(len(mapping1), len(mapping2))
            for (k1, v1), (k2, v2) in zip(
                sorted(mapping1.items()), sorted(mapping2.items())
            ):
                self.assertAlmostEqual(k1, k2)
                self.assertAlmostEqual(v1, v2)

    def test_compile(self):
        avar = table__a_v_a_r()
        avar.segments["wdth"] = {-1.0: -1.0, 0.0: 0.0, 0.3: 0.8, 1.0: 1.0}
        avar.segments["wght"] = {-1.0: -1.0, 0.0: 0.0, 1.0: 1.0}
        self.assertEqual(TEST_DATA, avar.compile(self.makeFont(["wdth", "wght"])))

    def test_decompile(self):
        avar = table__a_v_a_r()
        avar.decompile(TEST_DATA, self.makeFont(["wdth", "wght"]))
        self.assertAvarAlmostEqual(
            {
                "wdth": {-1.0: -1.0, 0.0: 0.0, 0.2999878: 0.7999878, 1.0: 1.0},
                "wght": {-1.0: -1.0, 0.0: 0.0, 1.0: 1.0},
            },
            avar.segments,
        )

    def test_toXML(self):
        avar = table__a_v_a_r()
        avar.segments["opsz"] = {-1.0: -1.0, 0.0: 0.0, 0.2999878: 0.7999878, 1.0: 1.0}
        writer = XMLWriter(BytesIO())
        avar.toXML(writer, self.makeFont(["opsz"]))
        self.assertEqual(
            [
                '<version major="1" minor="0"/>',
                '<segment axis="opsz">',
                '<mapping from="-1.0" to="-1.0"/>',
                '<mapping from="0.0" to="0.0"/>',
                '<mapping from="0.3" to="0.8"/>',
                '<mapping from="1.0" to="1.0"/>',
                "</segment>",
            ],
            self.xml_lines(writer),
        )

    def test_fromXML(self):
        avar = table__a_v_a_r()
        for name, attrs, content in parseXML(
            '<segment axis="wdth">'
            '    <mapping from="-1.0" to="-1.0"/>'
            '    <mapping from="0.0" to="0.0"/>'
            '    <mapping from="0.7" to="0.2"/>'
            '    <mapping from="1.0" to="1.0"/>'
            "</segment>"
        ):
            avar.fromXML(name, attrs, content, ttFont=None)
        self.assertAvarAlmostEqual(
            {"wdth": {-1: -1, 0: 0, 0.7000122: 0.2000122, 1.0: 1.0}}, avar.segments
        )

    @staticmethod
    def makeFont(axisTags):
        """['opsz', 'wdth'] --> ttFont"""
        fvar = table__f_v_a_r()
        for tag in axisTags:
            axis = Axis()
            axis.axisTag = tag
            fvar.axes.append(axis)
        font = TTFont()
        font["fvar"] = fvar
        return font

    @staticmethod
    def xml_lines(writer):
        content = writer.file.getvalue().decode("utf-8")
        return [line.strip() for line in content.splitlines()][1:]


class Avar2Test(unittest.TestCase):
    def test_roundtrip(self):
        axisTags = ["wght", "wdth"]
        fvar = table__f_v_a_r()
        for tag in axisTags:
            axis = Axis()
            axis.axisTag = tag
            fvar.axes.append(axis)

        master_locations_normalized = [
            {},
            {"wght": 1, "wdth": -1},
        ]
        data = [
            {},
            {"wdth": -0.8},
        ]

        model = models.VariationModel(master_locations_normalized, axisTags)
        store_builder = varStore.OnlineVarStoreBuilder(axisTags)
        store_builder.setModel(model)
        varIdxes = {}
        for axis in axisTags:
            masters = [fl2fi(m.get(axis, 0), 14) for m in data]
            varIdxes[axis] = store_builder.storeMasters(masters)[1]
        store = store_builder.finish()
        mapping = store.optimize()
        varIdxes = {axis: mapping[value] for axis, value in varIdxes.items()}
        del model, store_builder, mapping

        varIdxMap = otTables.DeltaSetIndexMap()
        varIdxMap.Format = 1
        varIdxMap.mapping = []
        for tag in axisTags:
            varIdxMap.mapping.append(varIdxes[tag])

        avar = table__a_v_a_r()
        avar.segments["wght"] = {}
        avar.segments["wdth"] = {-1.0: -1.0, 0.0: 0.0, 0.4: 0.5, 1.0: 1.0}

        avar.majorVersion = 2
        avar.table = otTables.avar()
        avar.table.VarIdxMap = varIdxMap
        avar.table.VarStore = store

        font = TTFont()
        font["fvar"] = fvar
        font["avar"] = avar

        b = BytesIO()
        font.save(b)
        b.seek(0)
        font2 = TTFont(b)

        assert font2["avar"].table.VarStore.VarRegionList.RegionAxisCount == 2
        assert font2["avar"].table.VarStore.VarRegionList.RegionCount == 1

        xml1 = BytesIO()
        writer = XMLWriter(xml1)
        font["avar"].toXML(writer, font)

        xml2 = BytesIO()
        writer = XMLWriter(xml2)
        font2["avar"].toXML(writer, font2)

        assert xml1.getvalue() == xml2.getvalue(), (xml1.getvalue(), xml2.getvalue())

        avar = table__a_v_a_r()
        xml = b"".join(xml2.getvalue().splitlines()[1:])
        for name, attrs, content in parseXML(xml):
            avar.fromXML(name, attrs, content, ttFont=TTFont())
        assert avar.table.VarStore.VarRegionList.RegionAxisCount == 2
        assert avar.table.VarStore.VarRegionList.RegionCount == 1

    def test_ttGlyphSet(self):
        ttf = os.path.join(DATA_DIR, "Amstelvar-avar2.subset.ttf")
        font = TTFont(ttf)

        regular = font.getGlyphSet()
        black = font.getGlyphSet(location={"wght": 900})
        expanded = font.getGlyphSet(location={"wdth": 125})

        regularStats = StatisticsPen()
        blackStats = StatisticsPen()
        expandedStats = StatisticsPen()

        for glyph in "Test":
            regular[glyph].draw(regularStats)
            black[glyph].draw(blackStats)
            expanded[glyph].draw(expandedStats)

        assert abs(regularStats.area) < abs(blackStats.area)
        assert abs(expandedStats.area) < abs(blackStats.area)

        assert regularStats.meanX < expandedStats.meanX


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



from .otBase import BaseTTXConverter


class table_S_T_A_T_(BaseTTXConverter):
    """Style Attributes table

    The ``STAT`` table records stylistic or typeface-design attributes that
    differentiate the individual fonts within a font family from one another.
    Those attributes can be used to assist users when navigating the style
    variations of a variable font or a family of static fonts.

    See also https://learn.microsoft.com/en-us/typography/opentype/spec/stat
    """

    pass



from fontTools.voltLib.error import VoltLibError
from fontTools.voltLib.lexer import Lexer
import unittest


def lex(s):
    return [(typ, tok) for (typ, tok, _) in Lexer(s, "test.vtp")]


class LexerTest(unittest.TestCase):
    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def test_empty(self):
        self.assertEqual(lex(""), [])
        self.assertEqual(lex("\t"), [])

    def test_string(self):
        self.assertEqual(
            lex('"foo" "bar"'), [(Lexer.STRING, "foo"), (Lexer.STRING, "bar")]
        )
        self.assertRaises(VoltLibError, lambda: lex('"foo\n bar"'))

    def test_name(self):
        self.assertEqual(
            lex("DEF_FOO bar.alt1"), [(Lexer.NAME, "DEF_FOO"), (Lexer.NAME, "bar.alt1")]
        )

    def test_number(self):
        self.assertEqual(lex("123 -456"), [(Lexer.NUMBER, 123), (Lexer.NUMBER, -456)])


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



from fontTools import ttLib

superclass = ttLib.getTableClass("fpgm")


class table__p_r_e_p(superclass):
    """Control Value Program table

    The ``prep`` table contains TrueType instructions that can makee font-wide
    alterations to the Control Value Table. It may potentially be executed
    before any glyph is processed.

    See also https://learn.microsoft.com/en-us/typography/opentype/spec/prep
    """

    pass



import fontTools.ttLib as ttLib
from fontTools.ttLib.ttVisitor import TTVisitor
from fontTools.misc.textTools import Tag
from array import array


class JsonVisitor(TTVisitor):
    def _open(self, s):
        print(s, file=self.file)
        self._indent += self.indent
        self.comma = False

    def _close(self, s):
        self._indent = self._indent[: -len(self.indent)]
        print("\n%s%s" % (self._indent, s), end="", file=self.file)
        self.comma = True

    def __init__(self, file, indent="  "):
        self.file = file
        self.indent = indent
        self._indent = ""

    def visitObject(self, obj):
        self._open("{")
        super().visitObject(obj)
        if self.comma:
            print(",", end="", file=self.file)
        print(
            '\n%s"type": "%s"' % (self._indent, obj.__class__.__name__),
            end="",
            file=self.file,
        )
        self._close("}")

    def visitAttr(self, obj, attr, value):
        if self.comma:
            print(",", file=self.file)
        print('%s"%s": ' % (self._indent, attr), end="", file=self.file)
        self.visit(value)
        self.comma = True

    def visitList(self, obj, *args, **kwargs):
        self._open("[")
        comma = False
        for value in obj:
            if comma:
                print(",", end="", file=self.file)
                print(file=self.file)
            print(self._indent, end="", file=self.file)
            self.visit(value, *args, **kwargs)
            comma = True
        self._close("]")

    def visitDict(self, obj, *args, **kwargs):
        self._open("{")
        comma = False
        for key, value in obj.items():
            if comma:
                print(",", end="", file=self.file)
                print(file=self.file)
            print('%s"%s": ' % (self._indent, key), end="", file=self.file)
            self.visit(value, *args, **kwargs)
            comma = True
        self._close("}")

    def visitLeaf(self, obj):
        if isinstance(obj, tuple):
            obj = list(obj)
        elif isinstance(obj, bytes):
            obj = list(obj)

        if obj is None:
            s = "null"
        elif obj is True:
            s = "true"
        elif obj is False:
            s = "false"
        else:
            s = repr(obj)

        if s[0] == "'":
            s = '"' + s[1:-1] + '"'

        print("%s" % s, end="", file=self.file)


@JsonVisitor.register(ttLib.TTFont)
def visit(self, font):
    if hasattr(visitor, "font"):
        print("{}", end="", file=self.file)
        return False
    visitor.font = font

    self._open("{")
    for tag in font.keys():
        if self.comma:
            print(",", file=self.file)
        print('\n%s"%s": ' % (self._indent, tag), end="", file=self.file)
        visitor.visit(font[tag])
    self._close("}")

    del visitor.font
    return False


@JsonVisitor.register(ttLib.GlyphOrder)
def visit(self, obj):
    self.visitList(self.font.getGlyphOrder())
    return False


@JsonVisitor.register_attr(ttLib.getTableClass("glyf"), "glyphOrder")
def visit(visitor, obj, attr, value):
    return False


@JsonVisitor.register(ttLib.getTableModule("glyf").GlyphCoordinates)
def visit(self, obj):
    self.visitList(obj)
    return False


@JsonVisitor.register(Tag)
def visit(self, obj):
    print('"%s"' % str(obj), end="", file=self.file)
    return False


@JsonVisitor.register(array)
def visit(self, obj):
    self.visitList(obj)
    return False


@JsonVisitor.register(bytearray)
def visit(self, obj):
    self.visitList(obj)
    return False


if __name__ == "__main__":
    from fontTools.ttLib import TTFont
    import sys

    if len(sys.argv) != 2:
        print("usage: print-json.py font")
        sys.exit()

    font = TTFont(sys.argv[1])

    visitor = JsonVisitor(sys.stdout)
    visitor.visit(font)



from .otBase import BaseTTXConverter


# https://developer.apple.com/fonts/TrueType-Reference-Manual/RM06/Chap6opbd.html
class table__o_p_b_d(BaseTTXConverter):
    """Optical Bounds table

    The AAT ``opbd`` table contains optical boundary points for glyphs, which
    applications can use for the visual alignment of lines of text.

    See also https://developer.apple.com/fonts/TrueType-Reference-Manual/RM06/Chap6opbd.html
    """

    pass



from fontTools.pens.t2CharStringPen import T2CharStringPen
import unittest


class T2CharStringPenTest(unittest.TestCase):
    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def assertAlmostEqualProgram(self, expected, actual):
        self.assertEqual(len(expected), len(actual))
        for i1, i2 in zip(expected, actual):
            if isinstance(i1, str):
                self.assertIsInstance(i2, str)
                self.assertEqual(i1, i2)
            else:
                self.assertAlmostEqual(i1, i2)

    def test_draw_h_v_lines(self):
        pen = T2CharStringPen(100, {})
        pen.moveTo((0, 0))
        pen.lineTo((10, 0))
        pen.lineTo((10, 10))
        pen.lineTo((0, 10))
        pen.closePath()  # no-op
        pen.moveTo((10, 10))
        pen.lineTo((10, 20))
        pen.lineTo((0, 20))
        pen.lineTo((0, 10))
        pen.closePath()
        charstring = pen.getCharString(None, None)

        self.assertEqual(
            [
                100,
                0,
                "hmoveto",
                10,
                10,
                -10,
                "hlineto",
                10,
                "hmoveto",
                10,
                -10,
                -10,
                "vlineto",
                "endchar",
            ],
            charstring.program,
        )

    def test_draw_lines(self):
        pen = T2CharStringPen(100, {})
        pen.moveTo((5, 5))
        pen.lineTo((25, 15))
        pen.lineTo((35, 35))
        pen.lineTo((15, 25))
        pen.closePath()  # no-op
        charstring = pen.getCharString(None, None)

        self.assertEqual(
            [100, 5, 5, "rmoveto", 20, 10, 10, 20, -20, -10, "rlineto", "endchar"],
            charstring.program,
        )

    def test_draw_h_v_curves(self):
        pen = T2CharStringPen(100, {})
        pen.moveTo((0, 0))
        pen.curveTo((10, 0), (20, 10), (20, 20))
        pen.curveTo((20, 30), (10, 40), (0, 40))
        pen.endPath()  # no-op
        charstring = pen.getCharString(None, None)

        self.assertEqual(
            [
                100,
                0,
                "hmoveto",
                10,
                10,
                10,
                10,
                10,
                -10,
                10,
                -10,
                "hvcurveto",
                "endchar",
            ],
            charstring.program,
        )

    def test_draw_curves(self):
        pen = T2CharStringPen(100, {})
        pen.moveTo((95, 25))
        pen.curveTo((115, 44), (115, 76), (95, 95))
        pen.curveTo((76, 114), (44, 115), (25, 95))
        pen.endPath()  # no-op
        charstring = pen.getCharString(None, None)

        self.assertEqual(
            [
                100,
                95,
                25,
                "rmoveto",
                20,
                19,
                0,
                32,
                -20,
                19,
                -19,
                19,
                -32,
                1,
                -19,
                -20,
                "rrcurveto",
                "endchar",
            ],
            charstring.program,
        )

    def test_draw_more_curves(self):
        pen = T2CharStringPen(100, {})
        pen.moveTo((10, 10))
        pen.curveTo((20, 10), (50, 10), (60, 10))
        pen.curveTo((60, 20), (60, 50), (60, 60))
        pen.curveTo((50, 50), (40, 60), (30, 60))
        pen.curveTo((40, 50), (30, 40), (30, 30))
        pen.curveTo((30, 25), (25, 19), (20, 20))
        pen.curveTo((15, 20), (9, 25), (10, 30))
        pen.curveTo((7, 25), (6, 15), (10, 10))
        pen.endPath()  # no-op
        charstring = pen.getCharString(None, None)

        self.assertEqual(
            [
                100,
                10,
                10,
                "rmoveto",
                10,
                30,
                0,
                10,
                "hhcurveto",
                10,
                0,
                30,
                10,
                "vvcurveto",
                -10,
                -10,
                -10,
                10,
                -10,
                "hhcurveto",
                10,
                -10,
                -10,
                -10,
                -10,
                "vvcurveto",
                -5,
                -5,
                -6,
                -5,
                1,
                "vhcurveto",
                -5,
                -6,
                5,
                5,
                1,
                "hvcurveto",
                -3,
                -5,
                -1,
                -10,
                4,
                -5,
                "rrcurveto",
                "endchar",
            ],
            charstring.program,
        )

    def test_default_width(self):
        pen = T2CharStringPen(None, {})
        charstring = pen.getCharString(None, None)
        self.assertEqual(["endchar"], charstring.program)

    def test_no_round(self):
        pen = T2CharStringPen(100.1, {}, roundTolerance=0.0)
        pen.moveTo((0, 0))
        pen.curveTo((10.1, 0.1), (19.9, 9.9), (20.49, 20.49))
        pen.curveTo((20.49, 30.49), (9.9, 39.9), (0.1, 40.1))
        pen.closePath()
        charstring = pen.getCharString(None, None)

        self.assertAlmostEqualProgram(
            [
                100,  # we always round the advance width
                0,
                "hmoveto",
                10.1,
                0.1,
                9.8,
                9.8,
                0.59,
                10.59,
                "rrcurveto",
                10,
                -10.59,
                9.41,
                -9.8,
                0.2,
                "vhcurveto",
                "endchar",
            ],
            charstring.program,
        )

    def test_round_all(self):
        pen = T2CharStringPen(100.1, {}, roundTolerance=0.5)
        pen.moveTo((0, 0))
        pen.curveTo((10.1, 0.1), (19.9, 9.9), (20.49, 20.49))
        pen.curveTo((20.49, 30.5), (9.9, 39.9), (0.1, 40.1))
        pen.closePath()
        charstring = pen.getCharString(None, None)

        self.assertEqual(
            [
                100,
                0,
                "hmoveto",
                10,
                10,
                10,
                10,
                11,
                -10,
                9,
                -10,
                "hvcurveto",
                "endchar",
            ],
            charstring.program,
        )

    def test_round_some(self):
        pen = T2CharStringPen(100, {}, roundTolerance=0.2)
        pen.moveTo((0, 0))
        # the following two are rounded as within the tolerance
        pen.lineTo((10.1, 0.1))
        pen.lineTo((19.9, 9.9))
        # this one is not rounded as it exceeds the tolerance
        pen.lineTo((20.49, 20.49))
        pen.closePath()
        charstring = pen.getCharString(None, None)

        self.assertAlmostEqualProgram(
            [
                100,
                0,
                "hmoveto",
                10,
                "hlineto",
                10,
                10,
                0.49,
                10.49,
                "rlineto",
                "endchar",
            ],
            charstring.program,
        )

    def test_invalid_tolerance(self):
        self.assertRaisesRegex(
            ValueError,
            "Rounding tolerance must be positive",
            T2CharStringPen,
            None,
            {},
            roundTolerance=-0.1,
        )


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



from copy import deepcopy
import string
from fontTools.colorLib.builder import LayerListBuilder, buildCOLR, buildClipList
from fontTools.misc.testTools import getXML
from fontTools.varLib.merger import COLRVariationMerger
from fontTools.varLib.models import VariationModel
from fontTools.ttLib import TTFont
from fontTools.ttLib.tables import otTables as ot
from fontTools.ttLib.tables.otBase import OTTableReader, OTTableWriter
from io import BytesIO
import pytest


NO_VARIATION_INDEX = ot.NO_VARIATION_INDEX


def dump_xml(table, ttFont=None):
    xml = getXML(table.toXML, ttFont)
    print("[")
    for line in xml:
        print(f"  {line!r},")
    print("]")
    return xml


def compile_decompile(table, ttFont):
    writer = OTTableWriter(tableTag="COLR")
    # compile itself may modify a table, safer to copy it first
    table = deepcopy(table)
    table.compile(writer, ttFont)
    data = writer.getAllData()

    reader = OTTableReader(data, tableTag="COLR")
    table2 = table.__class__()
    table2.decompile(reader, ttFont)

    return table2


@pytest.fixture
def ttFont():
    font = TTFont()
    font.setGlyphOrder([".notdef"] + list(string.ascii_letters))
    return font


def build_paint(data):
    return LayerListBuilder().buildPaint(data)


class COLRVariationMergerTest:
    @pytest.mark.parametrize(
        "paints, expected_xml, expected_varIdxes",
        [
            pytest.param(
                [
                    {
                        "Format": int(ot.PaintFormat.PaintSolid),
                        "PaletteIndex": 0,
                        "Alpha": 1.0,
                    },
                    {
                        "Format": int(ot.PaintFormat.PaintSolid),
                        "PaletteIndex": 0,
                        "Alpha": 1.0,
                    },
                ],
                [
                    '<Paint Format="2"><!-- PaintSolid -->',
                    '  <PaletteIndex value="0"/>',
                    '  <Alpha value="1.0"/>',
                    "</Paint>",
                ],
                [],
                id="solid-same",
            ),
            pytest.param(
                [
                    {
                        "Format": int(ot.PaintFormat.PaintSolid),
                        "PaletteIndex": 0,
                        "Alpha": 1.0,
                    },
                    {
                        "Format": int(ot.PaintFormat.PaintSolid),
                        "PaletteIndex": 0,
                        "Alpha": 0.5,
                    },
                ],
                [
                    '<Paint Format="3"><!-- PaintVarSolid -->',
                    '  <PaletteIndex value="0"/>',
                    '  <Alpha value="1.0"/>',
                    '  <VarIndexBase value="0"/>',
                    "</Paint>",
                ],
                [0],
                id="solid-alpha",
            ),
            pytest.param(
                [
                    {
                        "Format": int(ot.PaintFormat.PaintLinearGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.PAD),
                            "ColorStop": [
                                {"StopOffset": 0.0, "PaletteIndex": 0, "Alpha": 1.0},
                                {"StopOffset": 1.0, "PaletteIndex": 1, "Alpha": 1.0},
                            ],
                        },
                        "x0": 0,
                        "y0": 0,
                        "x1": 1,
                        "y1": 1,
                        "x2": 2,
                        "y2": 2,
                    },
                    {
                        "Format": int(ot.PaintFormat.PaintLinearGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.PAD),
                            "ColorStop": [
                                {"StopOffset": 0.1, "PaletteIndex": 0, "Alpha": 1.0},
                                {"StopOffset": 0.9, "PaletteIndex": 1, "Alpha": 1.0},
                            ],
                        },
                        "x0": 0,
                        "y0": 0,
                        "x1": 1,
                        "y1": 1,
                        "x2": 2,
                        "y2": 2,
                    },
                ],
                [
                    '<Paint Format="5"><!-- PaintVarLinearGradient -->',
                    "  <ColorLine>",
                    '    <Extend value="pad"/>',
                    "    <!-- StopCount=2 -->",
                    '    <ColorStop index="0">',
                    '      <StopOffset value="0.0"/>',
                    '      <PaletteIndex value="0"/>',
                    '      <Alpha value="1.0"/>',
                    '      <VarIndexBase value="0"/>',
                    "    </ColorStop>",
                    '    <ColorStop index="1">',
                    '      <StopOffset value="1.0"/>',
                    '      <PaletteIndex value="1"/>',
                    '      <Alpha value="1.0"/>',
                    '      <VarIndexBase value="2"/>',
                    "    </ColorStop>",
                    "  </ColorLine>",
                    '  <x0 value="0"/>',
                    '  <y0 value="0"/>',
                    '  <x1 value="1"/>',
                    '  <y1 value="1"/>',
                    '  <x2 value="2"/>',
                    '  <y2 value="2"/>',
                    "  <VarIndexBase/>",
                    "</Paint>",
                ],
                [0, NO_VARIATION_INDEX, 1, NO_VARIATION_INDEX],
                id="linear_grad-stop-offsets",
            ),
            pytest.param(
                [
                    {
                        "Format": int(ot.PaintFormat.PaintLinearGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.PAD),
                            "ColorStop": [
                                {"StopOffset": 0.0, "PaletteIndex": 0, "Alpha": 1.0},
                                {"StopOffset": 1.0, "PaletteIndex": 1, "Alpha": 1.0},
                            ],
                        },
                        "x0": 0,
                        "y0": 0,
                        "x1": 1,
                        "y1": 1,
                        "x2": 2,
                        "y2": 2,
                    },
                    {
                        "Format": int(ot.PaintFormat.PaintLinearGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.PAD),
                            "ColorStop": [
                                {"StopOffset": 0.0, "PaletteIndex": 0, "Alpha": 0.5},
                                {"StopOffset": 1.0, "PaletteIndex": 1, "Alpha": 1.0},
                            ],
                        },
                        "x0": 0,
                        "y0": 0,
                        "x1": 1,
                        "y1": 1,
                        "x2": 2,
                        "y2": 2,
                    },
                ],
                [
                    '<Paint Format="5"><!-- PaintVarLinearGradient -->',
                    "  <ColorLine>",
                    '    <Extend value="pad"/>',
                    "    <!-- StopCount=2 -->",
                    '    <ColorStop index="0">',
                    '      <StopOffset value="0.0"/>',
                    '      <PaletteIndex value="0"/>',
                    '      <Alpha value="1.0"/>',
                    '      <VarIndexBase value="0"/>',
                    "    </ColorStop>",
                    '    <ColorStop index="1">',
                    '      <StopOffset value="1.0"/>',
                    '      <PaletteIndex value="1"/>',
                    '      <Alpha value="1.0"/>',
                    "      <VarIndexBase/>",
                    "    </ColorStop>",
                    "  </ColorLine>",
                    '  <x0 value="0"/>',
                    '  <y0 value="0"/>',
                    '  <x1 value="1"/>',
                    '  <y1 value="1"/>',
                    '  <x2 value="2"/>',
                    '  <y2 value="2"/>',
                    "  <VarIndexBase/>",
                    "</Paint>",
                ],
                [NO_VARIATION_INDEX, 0],
                id="linear_grad-stop[0].alpha",
            ),
            pytest.param(
                [
                    {
                        "Format": int(ot.PaintFormat.PaintLinearGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.PAD),
                            "ColorStop": [
                                {"StopOffset": 0.0, "PaletteIndex": 0, "Alpha": 1.0},
                                {"StopOffset": 1.0, "PaletteIndex": 1, "Alpha": 1.0},
                            ],
                        },
                        "x0": 0,
                        "y0": 0,
                        "x1": 1,
                        "y1": 1,
                        "x2": 2,
                        "y2": 2,
                    },
                    {
                        "Format": int(ot.PaintFormat.PaintLinearGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.PAD),
                            "ColorStop": [
                                {"StopOffset": -0.5, "PaletteIndex": 0, "Alpha": 1.0},
                                {"StopOffset": 1.0, "PaletteIndex": 1, "Alpha": 1.0},
                            ],
                        },
                        "x0": 0,
                        "y0": 0,
                        "x1": 1,
                        "y1": 1,
                        "x2": 2,
                        "y2": -200,
                    },
                ],
                [
                    '<Paint Format="5"><!-- PaintVarLinearGradient -->',
                    "  <ColorLine>",
                    '    <Extend value="pad"/>',
                    "    <!-- StopCount=2 -->",
                    '    <ColorStop index="0">',
                    '      <StopOffset value="0.0"/>',
                    '      <PaletteIndex value="0"/>',
                    '      <Alpha value="1.0"/>',
                    '      <VarIndexBase value="0"/>',
                    "    </ColorStop>",
                    '    <ColorStop index="1">',
                    '      <StopOffset value="1.0"/>',
                    '      <PaletteIndex value="1"/>',
                    '      <Alpha value="1.0"/>',
                    "      <VarIndexBase/>",
                    "    </ColorStop>",
                    "  </ColorLine>",
                    '  <x0 value="0"/>',
                    '  <y0 value="0"/>',
                    '  <x1 value="1"/>',
                    '  <y1 value="1"/>',
                    '  <x2 value="2"/>',
                    '  <y2 value="2"/>',
                    '  <VarIndexBase value="1"/>',
                    "</Paint>",
                ],
                [
                    0,
                    NO_VARIATION_INDEX,
                    NO_VARIATION_INDEX,
                    NO_VARIATION_INDEX,
                    NO_VARIATION_INDEX,
                    NO_VARIATION_INDEX,
                    1,
                ],
                id="linear_grad-stop[0].offset-y2",
            ),
            pytest.param(
                [
                    {
                        "Format": int(ot.PaintFormat.PaintRadialGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.PAD),
                            "ColorStop": [
                                {"StopOffset": 0.0, "PaletteIndex": 0, "Alpha": 1.0},
                                {"StopOffset": 1.0, "PaletteIndex": 1, "Alpha": 1.0},
                            ],
                        },
                        "x0": 0,
                        "y0": 0,
                        "r0": 0,
                        "x1": 1,
                        "y1": 1,
                        "r1": 1,
                    },
                    {
                        "Format": int(ot.PaintFormat.PaintRadialGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.PAD),
                            "ColorStop": [
                                {"StopOffset": 0.1, "PaletteIndex": 0, "Alpha": 0.6},
                                {"StopOffset": 0.9, "PaletteIndex": 1, "Alpha": 0.7},
                            ],
                        },
                        "x0": -1,
                        "y0": -2,
                        "r0": 3,
                        "x1": -4,
                        "y1": -5,
                        "r1": 6,
                    },
                ],
                [
                    '<Paint Format="7"><!-- PaintVarRadialGradient -->',
                    "  <ColorLine>",
                    '    <Extend value="pad"/>',
                    "    <!-- StopCount=2 -->",
                    '    <ColorStop index="0">',
                    '      <StopOffset value="0.0"/>',
                    '      <PaletteIndex value="0"/>',
                    '      <Alpha value="1.0"/>',
                    '      <VarIndexBase value="0"/>',
                    "    </ColorStop>",
                    '    <ColorStop index="1">',
                    '      <StopOffset value="1.0"/>',
                    '      <PaletteIndex value="1"/>',
                    '      <Alpha value="1.0"/>',
                    '      <VarIndexBase value="2"/>',
                    "    </ColorStop>",
                    "  </ColorLine>",
                    '  <x0 value="0"/>',
                    '  <y0 value="0"/>',
                    '  <r0 value="0"/>',
                    '  <x1 value="1"/>',
                    '  <y1 value="1"/>',
                    '  <r1 value="1"/>',
                    '  <VarIndexBase value="4"/>',
                    "</Paint>",
                ],
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                id="radial_grad-all-different",
            ),
            pytest.param(
                [
                    {
                        "Format": int(ot.PaintFormat.PaintSweepGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.REPEAT),
                            "ColorStop": [
                                {"StopOffset": 0.4, "PaletteIndex": 0, "Alpha": 1.0},
                                {"StopOffset": 0.6, "PaletteIndex": 1, "Alpha": 1.0},
                            ],
                        },
                        "centerX": 0,
                        "centerY": 0,
                        "startAngle": 0,
                        "endAngle": 180.0,
                    },
                    {
                        "Format": int(ot.PaintFormat.PaintSweepGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.REPEAT),
                            "ColorStop": [
                                {"StopOffset": 0.4, "PaletteIndex": 0, "Alpha": 1.0},
                                {"StopOffset": 0.6, "PaletteIndex": 1, "Alpha": 1.0},
                            ],
                        },
                        "centerX": 0,
                        "centerY": 0,
                        "startAngle": 90.0,
                        "endAngle": 180.0,
                    },
                ],
                [
                    '<Paint Format="9"><!-- PaintVarSweepGradient -->',
                    "  <ColorLine>",
                    '    <Extend value="repeat"/>',
                    "    <!-- StopCount=2 -->",
                    '    <ColorStop index="0">',
                    '      <StopOffset value="0.4"/>',
                    '      <PaletteIndex value="0"/>',
                    '      <Alpha value="1.0"/>',
                    "      <VarIndexBase/>",
                    "    </ColorStop>",
                    '    <ColorStop index="1">',
                    '      <StopOffset value="0.6"/>',
                    '      <PaletteIndex value="1"/>',
                    '      <Alpha value="1.0"/>',
                    "      <VarIndexBase/>",
                    "    </ColorStop>",
                    "  </ColorLine>",
                    '  <centerX value="0"/>',
                    '  <centerY value="0"/>',
                    '  <startAngle value="0.0"/>',
                    '  <endAngle value="180.0"/>',
                    '  <VarIndexBase value="0"/>',
                    "</Paint>",
                ],
                [NO_VARIATION_INDEX, NO_VARIATION_INDEX, 0, NO_VARIATION_INDEX],
                id="sweep_grad-startAngle",
            ),
            pytest.param(
                [
                    {
                        "Format": int(ot.PaintFormat.PaintSweepGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.PAD),
                            "ColorStop": [
                                {"StopOffset": 0.0, "PaletteIndex": 0, "Alpha": 1.0},
                                {"StopOffset": 1.0, "PaletteIndex": 1, "Alpha": 1.0},
                            ],
                        },
                        "centerX": 0,
                        "centerY": 0,
                        "startAngle": 0.0,
                        "endAngle": 180.0,
                    },
                    {
                        "Format": int(ot.PaintFormat.PaintSweepGradient),
                        "ColorLine": {
                            "Extend": int(ot.ExtendMode.PAD),
                            "ColorStop": [
                                {"StopOffset": 0.0, "PaletteIndex": 0, "Alpha": 0.5},
                                {"StopOffset": 1.0, "PaletteIndex": 1, "Alpha": 0.5},
                            ],
                        },
                        "centerX": 0,
                        "centerY": 0,
                        "startAngle": 0.0,
                        "endAngle": 180.0,
                    },
                ],
                [
                    '<Paint Format="9"><!-- PaintVarSweepGradient -->',
                    "  <ColorLine>",
                    '    <Extend value="pad"/>',
                    "    <!-- StopCount=2 -->",
                    '    <ColorStop index="0">',
                    '      <StopOffset value="0.0"/>',
                    '      <PaletteIndex value="0"/>',
                    '      <Alpha value="1.0"/>',
                    '      <VarIndexBase value="0"/>',
                    "    </ColorStop>",
                    '    <ColorStop index="1">',
                    '      <StopOffset value="1.0"/>',
                    '      <PaletteIndex value="1"/>',
                    '      <Alpha value="1.0"/>',
                    '      <VarIndexBase value="0"/>',
                    "    </ColorStop>",
                    "  </ColorLine>",
                    '  <centerX value="0"/>',
                    '  <centerY value="0"/>',
                    '  <startAngle value="0.0"/>',
                    '  <endAngle value="180.0"/>',
                    "  <VarIndexBase/>",
                    "</Paint>",
                ],
                [NO_VARIATION_INDEX, 0],
                id="sweep_grad-stops-alpha-reuse-varidxbase",
            ),
            pytest.param(
                [
                    {
                        "Format": int(ot.PaintFormat.PaintTransform),
                        "Paint": {
                            "Format": int(ot.PaintFormat.PaintRadialGradient),
                            "ColorLine": {
                                "Extend": int(ot.ExtendMode.PAD),
                                "ColorStop": [
                                    {
                                        "StopOffset": 0.0,
                                        "PaletteIndex": 0,
                                        "Alpha": 1.0,
                                    },
                                    {
                                        "StopOffset": 1.0,
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                ],
                            },
                            "x0": 0,
                            "y0": 0,
                            "r0": 0,
                            "x1": 1,
                            "y1": 1,
                            "r1": 1,
                        },
                        "Transform": {
                            "xx": 1.0,
                            "xy": 0.0,
                            "yx": 0.0,
                            "yy": 1.0,
                            "dx": 0.0,
                            "dy": 0.0,
                        },
                    },
                    {
                        "Format": int(ot.PaintFormat.PaintTransform),
                        "Paint": {
                            "Format": int(ot.PaintFormat.PaintRadialGradient),
                            "ColorLine": {
                                "Extend": int(ot.ExtendMode.PAD),
                                "ColorStop": [
                                    {
                                        "StopOffset": 0.0,
                                        "PaletteIndex": 0,
                                        "Alpha": 1.0,
                                    },
                                    {
                                        "StopOffset": 1.0,
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                ],
                            },
                            "x0": 0,
                            "y0": 0,
                            "r0": 0,
                            "x1": 1,
                            "y1": 1,
                            "r1": 1,
                        },
                        "Transform": {
                            "xx": 1.0,
                            "xy": 0.0,
                            "yx": 0.0,
                            "yy": 0.5,
                            "dx": 0.0,
                            "dy": -100.0,
                        },
                    },
                ],
                [
                    '<Paint Format="13"><!-- PaintVarTransform -->',
                    '  <Paint Format="6"><!-- PaintRadialGradient -->',
                    "    <ColorLine>",
                    '      <Extend value="pad"/>',
                    "      <!-- StopCount=2 -->",
                    '      <ColorStop index="0">',
                    '        <StopOffset value="0.0"/>',
                    '        <PaletteIndex value="0"/>',
                    '        <Alpha value="1.0"/>',
                    "      </ColorStop>",
                    '      <ColorStop index="1">',
                    '        <StopOffset value="1.0"/>',
                    '        <PaletteIndex value="1"/>',
                    '        <Alpha value="1.0"/>',
                    "      </ColorStop>",
                    "    </ColorLine>",
                    '    <x0 value="0"/>',
                    '    <y0 value="0"/>',
                    '    <r0 value="0"/>',
                    '    <x1 value="1"/>',
                    '    <y1 value="1"/>',
                    '    <r1 value="1"/>',
                    "  </Paint>",
                    "  <Transform>",
                    '    <xx value="1.0"/>',
                    '    <yx value="0.0"/>',
                    '    <xy value="0.0"/>',
                    '    <yy value="1.0"/>',
                    '    <dx value="0.0"/>',
                    '    <dy value="0.0"/>',
                    '    <VarIndexBase value="0"/>',
                    "  </Transform>",
                    "</Paint>",
                ],
                [
                    NO_VARIATION_INDEX,
                    NO_VARIATION_INDEX,
                    NO_VARIATION_INDEX,
                    0,
                    NO_VARIATION_INDEX,
                    1,
                ],
                id="transform-yy-dy",
            ),
            pytest.param(
                [
                    {
                        "Format": ot.PaintFormat.PaintTransform,
                        "Paint": {
                            "Format": ot.PaintFormat.PaintSweepGradient,
                            "ColorLine": {
                                "Extend": ot.ExtendMode.PAD,
                                "ColorStop": [
                                    {"StopOffset": 0.0, "PaletteIndex": 0},
                                    {
                                        "StopOffset": 1.0,
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                ],
                            },
                            "centerX": 0,
                            "centerY": 0,
                            "startAngle": 0,
                            "endAngle": 360,
                        },
                        "Transform": (1.0, 0, 0, 1.0, 0, 0),
                    },
                    {
                        "Format": ot.PaintFormat.PaintTransform,
                        "Paint": {
                            "Format": ot.PaintFormat.PaintSweepGradient,
                            "ColorLine": {
                                "Extend": ot.ExtendMode.PAD,
                                "ColorStop": [
                                    {"StopOffset": 0.0, "PaletteIndex": 0},
                                    {
                                        "StopOffset": 1.0,
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                ],
                            },
                            "centerX": 256,
                            "centerY": 0,
                            "startAngle": 0,
                            "endAngle": 360,
                        },
                        # Transform.xx below produces the same VarStore delta as the
                        # above PaintSweepGradient's centerX because, when Fixed16.16
                        # is converted to integer, it becomes:
                        # floatToFixed(1.00390625, 16) == 256
                        # Because there is overlap between the varIdxes of the
                        # PaintVarTransform's Affine2x3 and the PaintSweepGradient's
                        # the VarIndexBase is reused (0 for both)
                        "Transform": (1.00390625, 0, 0, 1.0, 10, 0),
                    },
                ],
                [
                    '<Paint Format="13"><!-- PaintVarTransform -->',
                    '  <Paint Format="9"><!-- PaintVarSweepGradient -->',
                    "    <ColorLine>",
                    '      <Extend value="pad"/>',
                    "      <!-- StopCount=2 -->",
                    '      <ColorStop index="0">',
                    '        <StopOffset value="0.0"/>',
                    '        <PaletteIndex value="0"/>',
                    '        <Alpha value="1.0"/>',
                    "        <VarIndexBase/>",
                    "      </ColorStop>",
                    '      <ColorStop index="1">',
                    '        <StopOffset value="1.0"/>',
                    '        <PaletteIndex value="1"/>',
                    '        <Alpha value="1.0"/>',
                    "        <VarIndexBase/>",
                    "      </ColorStop>",
                    "    </ColorLine>",
                    '    <centerX value="0"/>',
                    '    <centerY value="0"/>',
                    '    <startAngle value="0.0"/>',
                    '    <endAngle value="360.0"/>',
                    '    <VarIndexBase value="0"/>',
                    "  </Paint>",
                    "  <Transform>",
                    '    <xx value="1.0"/>',
                    '    <yx value="0.0"/>',
                    '    <xy value="0.0"/>',
                    '    <yy value="1.0"/>',
                    '    <dx value="0.0"/>',
                    '    <dy value="0.0"/>',
                    '    <VarIndexBase value="0"/>',
                    "  </Transform>",
                    "</Paint>",
                ],
                [
                    0,
                    NO_VARIATION_INDEX,
                    NO_VARIATION_INDEX,
                    NO_VARIATION_INDEX,
                    1,
                    NO_VARIATION_INDEX,
                ],
                id="transform-xx-sweep_grad-centerx-same-varidxbase",
            ),
        ],
    )
    def test_merge_Paint(self, paints, ttFont, expected_xml, expected_varIdxes):
        paints = [build_paint(p) for p in paints]
        out = deepcopy(paints[0])

        model = VariationModel([{}, {"ZZZZ": 1.0}])
        merger = COLRVariationMerger(model, ["ZZZZ"], ttFont)

        merger.mergeThings(out, paints)

        assert compile_decompile(out, ttFont) == out
        assert dump_xml(out, ttFont) == expected_xml
        assert merger.varIdxes == expected_varIdxes

    def test_merge_ClipList(self, ttFont):
        clipLists = [
            buildClipList(clips)
            for clips in [
                {
                    "A": (0, 0, 1000, 1000),
                    "B": (0, 0, 1000, 1000),
                    "C": (0, 0, 1000, 1000),
                    "D": (0, 0, 1000, 1000),
                },
                {
                    # non-default masters' clip boxes can be 'sparse'
                    # (i.e. can omit explicit clip box for some glyphs)
                    # "A": (0, 0, 1000, 1000),
                    "B": (10, 0, 1000, 1000),
                    "C": (20, 20, 1020, 1020),
                    "D": (20, 20, 1020, 1020),
                },
            ]
        ]
        out = deepcopy(clipLists[0])

        model = VariationModel([{}, {"ZZZZ": 1.0}])
        merger = COLRVariationMerger(model, ["ZZZZ"], ttFont)

        merger.mergeThings(out, clipLists)

        assert compile_decompile(out, ttFont) == out
        assert dump_xml(out, ttFont) == [
            '<ClipList Format="1">',
            "  <Clip>",
            '    <Glyph value="A"/>',
            '    <ClipBox Format="1">',
            '      <xMin value="0"/>',
            '      <yMin value="0"/>',
            '      <xMax value="1000"/>',
            '      <yMax value="1000"/>',
            "    </ClipBox>",
            "  </Clip>",
            "  <Clip>",
            '    <Glyph value="B"/>',
            '    <ClipBox Format="2">',
            '      <xMin value="0"/>',
            '      <yMin value="0"/>',
            '      <xMax value="1000"/>',
            '      <yMax value="1000"/>',
            '      <VarIndexBase value="0"/>',
            "    </ClipBox>",
            "  </Clip>",
            "  <Clip>",
            '    <Glyph value="C"/>',
            '    <Glyph value="D"/>',
            '    <ClipBox Format="2">',
            '      <xMin value="0"/>',
            '      <yMin value="0"/>',
            '      <xMax value="1000"/>',
            '      <yMax value="1000"/>',
            '      <VarIndexBase value="4"/>',
            "    </ClipBox>",
            "  </Clip>",
            "</ClipList>",
        ]
        assert merger.varIdxes == [
            0,
            NO_VARIATION_INDEX,
            NO_VARIATION_INDEX,
            NO_VARIATION_INDEX,
            1,
            1,
            1,
            1,
        ]

    @pytest.mark.parametrize(
        "master_layer_reuse",
        [
            pytest.param(False, id="no-reuse"),
            pytest.param(True, id="with-reuse"),
        ],
    )
    @pytest.mark.parametrize(
        "color_glyphs, output_layer_reuse, expected_xml, expected_varIdxes",
        [
            pytest.param(
                [
                    {
                        "A": {
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 0,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                    },
                    {
                        "A": {
                            "Format": ot.PaintFormat.PaintColrLayers,
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 0,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                    },
                ],
                False,
                [
                    "<COLR>",
                    '  <Version value="1"/>',
                    "  <!-- BaseGlyphRecordCount=0 -->",
                    "  <!-- LayerRecordCount=0 -->",
                    "  <BaseGlyphList>",
                    "    <!-- BaseGlyphCount=1 -->",
                    '    <BaseGlyphPaintRecord index="0">',
                    '      <BaseGlyph value="A"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="2"/>',
                    '        <FirstLayerIndex value="0"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    "  </BaseGlyphList>",
                    "  <LayerList>",
                    "    <!-- LayerCount=2 -->",
                    '    <Paint index="0" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="0"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="1" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="1"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    "  </LayerList>",
                    "</COLR>",
                ],
                [],
                id="no-variation",
            ),
            pytest.param(
                [
                    {
                        "A": {
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 0,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                        "C": {
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 3,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                    },
                    {
                        # NOTE: 'A' is missing from non-default master
                        "C": {
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 0.5,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 3,
                                        "Alpha": 0.5,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                    },
                ],
                False,
                [
                    "<COLR>",
                    '  <Version value="1"/>',
                    "  <!-- BaseGlyphRecordCount=0 -->",
                    "  <!-- LayerRecordCount=0 -->",
                    "  <BaseGlyphList>",
                    "    <!-- BaseGlyphCount=2 -->",
                    '    <BaseGlyphPaintRecord index="0">',
                    '      <BaseGlyph value="A"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="2"/>',
                    '        <FirstLayerIndex value="0"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    '    <BaseGlyphPaintRecord index="1">',
                    '      <BaseGlyph value="C"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="2"/>',
                    '        <FirstLayerIndex value="2"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    "  </BaseGlyphList>",
                    "  <LayerList>",
                    "    <!-- LayerCount=4 -->",
                    '    <Paint index="0" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="0"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="1" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="1"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="2" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="3"><!-- PaintVarSolid -->',
                    '        <PaletteIndex value="2"/>',
                    '        <Alpha value="1.0"/>',
                    '        <VarIndexBase value="0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="3" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="3"><!-- PaintVarSolid -->',
                    '        <PaletteIndex value="3"/>',
                    '        <Alpha value="1.0"/>',
                    '        <VarIndexBase value="0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    "  </LayerList>",
                    "</COLR>",
                ],
                [0],
                id="sparse-masters",
            ),
            pytest.param(
                [
                    {
                        "A": {
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 0,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                        "C": {
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                # 'C' reuses layers 1-3 from 'A'
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                        "D": {  # identical to 'C'
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                        "E": {  # superset of 'C' or 'D'
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 3,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                    },
                    {
                        # NOTE: 'A' is missing from non-default master
                        "C": {
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 0.5,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 0.5,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                        "D": {  # same as 'C'
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 0.5,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 0.5,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                        "E": {  # first two layers vary the same way as 'C' or 'D'
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 0.5,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 0.5,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 3,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                    },
                ],
                True,  # reuse
                [
                    "<COLR>",
                    '  <Version value="1"/>',
                    "  <!-- BaseGlyphRecordCount=0 -->",
                    "  <!-- LayerRecordCount=0 -->",
                    "  <BaseGlyphList>",
                    "    <!-- BaseGlyphCount=4 -->",
                    '    <BaseGlyphPaintRecord index="0">',
                    '      <BaseGlyph value="A"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="3"/>',
                    '        <FirstLayerIndex value="0"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    '    <BaseGlyphPaintRecord index="1">',
                    '      <BaseGlyph value="C"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="2"/>',
                    '        <FirstLayerIndex value="3"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    '    <BaseGlyphPaintRecord index="2">',
                    '      <BaseGlyph value="D"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="2"/>',
                    '        <FirstLayerIndex value="3"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    '    <BaseGlyphPaintRecord index="3">',
                    '      <BaseGlyph value="E"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="2"/>',
                    '        <FirstLayerIndex value="5"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    "  </BaseGlyphList>",
                    "  <LayerList>",
                    "    <!-- LayerCount=7 -->",
                    '    <Paint index="0" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="0"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="1" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="1"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="2" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="2"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="3" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="3"><!-- PaintVarSolid -->',
                    '        <PaletteIndex value="1"/>',
                    '        <Alpha value="1.0"/>',
                    '        <VarIndexBase value="0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="4" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="3"><!-- PaintVarSolid -->',
                    '        <PaletteIndex value="2"/>',
                    '        <Alpha value="1.0"/>',
                    '        <VarIndexBase value="0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="5" Format="1"><!-- PaintColrLayers -->',
                    '      <NumLayers value="2"/>',
                    '      <FirstLayerIndex value="3"/>',
                    "    </Paint>",
                    '    <Paint index="6" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="3"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    "  </LayerList>",
                    "</COLR>",
                ],
                [0],
                id="sparse-masters-with-reuse",
            ),
            pytest.param(
                [
                    {
                        "A": {
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 0,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                        "C": {  # 'C' shares layer 1 and 2 with 'A'
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                    },
                    {
                        "A": {
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 0,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 0.9,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                        "C": {
                            "Format": int(ot.PaintFormat.PaintColrLayers),
                            "Layers": [
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 1,
                                        "Alpha": 0.5,
                                    },
                                    "Glyph": "B",
                                },
                                {
                                    "Format": int(ot.PaintFormat.PaintGlyph),
                                    "Paint": {
                                        "Format": int(ot.PaintFormat.PaintSolid),
                                        "PaletteIndex": 2,
                                        "Alpha": 1.0,
                                    },
                                    "Glyph": "B",
                                },
                            ],
                        },
                    },
                ],
                True,
                [
                    # a different Alpha variation is applied to a shared layer between
                    # 'A' and 'C' and thus they are no longer shared.
                    "<COLR>",
                    '  <Version value="1"/>',
                    "  <!-- BaseGlyphRecordCount=0 -->",
                    "  <!-- LayerRecordCount=0 -->",
                    "  <BaseGlyphList>",
                    "    <!-- BaseGlyphCount=2 -->",
                    '    <BaseGlyphPaintRecord index="0">',
                    '      <BaseGlyph value="A"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="3"/>',
                    '        <FirstLayerIndex value="0"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    '    <BaseGlyphPaintRecord index="1">',
                    '      <BaseGlyph value="C"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="2"/>',
                    '        <FirstLayerIndex value="3"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    "  </BaseGlyphList>",
                    "  <LayerList>",
                    "    <!-- LayerCount=5 -->",
                    '    <Paint index="0" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="0"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="1" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="3"><!-- PaintVarSolid -->',
                    '        <PaletteIndex value="1"/>',
                    '        <Alpha value="1.0"/>',
                    '        <VarIndexBase value="0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="2" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="2"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="3" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="3"><!-- PaintVarSolid -->',
                    '        <PaletteIndex value="1"/>',
                    '        <Alpha value="1.0"/>',
                    '        <VarIndexBase value="1"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="4" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="2"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    "  </LayerList>",
                    "</COLR>",
                ],
                [0, 1],
                id="shared-master-layers-different-variations",
            ),
        ],
    )
    def test_merge_full_table(
        self,
        color_glyphs,
        ttFont,
        expected_xml,
        expected_varIdxes,
        master_layer_reuse,
        output_layer_reuse,
    ):
        master_ttfs = [deepcopy(ttFont) for _ in range(len(color_glyphs))]
        for ttf, glyphs in zip(master_ttfs, color_glyphs):
            # merge algorithm is expected to work the same even if the master COLRs
            # may differ as to the layer reuse, hence we try both ways
            ttf["COLR"] = buildCOLR(glyphs, allowLayerReuse=master_layer_reuse)
        vf = deepcopy(master_ttfs[0])

        model = VariationModel([{}, {"ZZZZ": 1.0}])
        merger = COLRVariationMerger(
            model, ["ZZZZ"], vf, allowLayerReuse=output_layer_reuse
        )

        merger.mergeTables(vf, master_ttfs)

        out = vf["COLR"].table

        assert compile_decompile(out, vf) == out
        assert dump_xml(out, vf) == expected_xml
        assert merger.varIdxes == expected_varIdxes

    @pytest.mark.parametrize(
        "color_glyphs, before_xml, expected_xml",
        [
            pytest.param(
                {
                    "A": {
                        "Format": int(ot.PaintFormat.PaintColrLayers),
                        "Layers": [
                            {
                                "Format": int(ot.PaintFormat.PaintGlyph),
                                "Paint": {
                                    "Format": int(ot.PaintFormat.PaintSolid),
                                    "PaletteIndex": 0,
                                    "Alpha": 1.0,
                                },
                                "Glyph": "B",
                            },
                            {
                                "Format": int(ot.PaintFormat.PaintGlyph),
                                "Paint": {
                                    "Format": int(ot.PaintFormat.PaintSolid),
                                    "PaletteIndex": 1,
                                    "Alpha": 1.0,
                                },
                                "Glyph": "C",
                            },
                            {
                                "Format": int(ot.PaintFormat.PaintGlyph),
                                "Paint": {
                                    "Format": int(ot.PaintFormat.PaintSolid),
                                    "PaletteIndex": 2,
                                    "Alpha": 1.0,
                                },
                                "Glyph": "D",
                            },
                        ],
                    },
                    "E": {
                        "Format": int(ot.PaintFormat.PaintColrLayers),
                        "Layers": [
                            {
                                "Format": int(ot.PaintFormat.PaintGlyph),
                                "Paint": {
                                    "Format": int(ot.PaintFormat.PaintSolid),
                                    "PaletteIndex": 1,
                                    "Alpha": 1.0,
                                },
                                "Glyph": "C",
                            },
                            {
                                "Format": int(ot.PaintFormat.PaintGlyph),
                                "Paint": {
                                    "Format": int(ot.PaintFormat.PaintSolid),
                                    "PaletteIndex": 2,
                                    "Alpha": 1.0,
                                },
                                "Glyph": "D",
                            },
                            {
                                "Format": int(ot.PaintFormat.PaintGlyph),
                                "Paint": {
                                    "Format": int(ot.PaintFormat.PaintSolid),
                                    "PaletteIndex": 3,
                                    "Alpha": 1.0,
                                },
                                "Glyph": "F",
                            },
                        ],
                    },
                    "G": {
                        "Format": int(ot.PaintFormat.PaintColrGlyph),
                        "Glyph": "E",
                    },
                },
                [
                    "<COLR>",
                    '  <Version value="1"/>',
                    "  <!-- BaseGlyphRecordCount=0 -->",
                    "  <!-- LayerRecordCount=0 -->",
                    "  <BaseGlyphList>",
                    "    <!-- BaseGlyphCount=3 -->",
                    '    <BaseGlyphPaintRecord index="0">',
                    '      <BaseGlyph value="A"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="3"/>',
                    '        <FirstLayerIndex value="0"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    '    <BaseGlyphPaintRecord index="1">',
                    '      <BaseGlyph value="E"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="2"/>',
                    '        <FirstLayerIndex value="3"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    '    <BaseGlyphPaintRecord index="2">',
                    '      <BaseGlyph value="G"/>',
                    '      <Paint Format="11"><!-- PaintColrGlyph -->',
                    '        <Glyph value="E"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    "  </BaseGlyphList>",
                    "  <LayerList>",
                    "    <!-- LayerCount=5 -->",
                    '    <Paint index="0" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="0"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="1" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="1"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="C"/>',
                    "    </Paint>",
                    '    <Paint index="2" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="2"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="D"/>',
                    "    </Paint>",
                    '    <Paint index="3" Format="1"><!-- PaintColrLayers -->',
                    '      <NumLayers value="2"/>',
                    '      <FirstLayerIndex value="1"/>',
                    "    </Paint>",
                    '    <Paint index="4" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="3"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="F"/>',
                    "    </Paint>",
                    "  </LayerList>",
                    "</COLR>",
                ],
                [
                    "<COLR>",
                    '  <Version value="1"/>',
                    "  <!-- BaseGlyphRecordCount=0 -->",
                    "  <!-- LayerRecordCount=0 -->",
                    "  <BaseGlyphList>",
                    "    <!-- BaseGlyphCount=3 -->",
                    '    <BaseGlyphPaintRecord index="0">',
                    '      <BaseGlyph value="A"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="3"/>',
                    '        <FirstLayerIndex value="0"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    '    <BaseGlyphPaintRecord index="1">',
                    '      <BaseGlyph value="E"/>',
                    '      <Paint Format="1"><!-- PaintColrLayers -->',
                    '        <NumLayers value="3"/>',
                    '        <FirstLayerIndex value="3"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    '    <BaseGlyphPaintRecord index="2">',
                    '      <BaseGlyph value="G"/>',
                    '      <Paint Format="11"><!-- PaintColrGlyph -->',
                    '        <Glyph value="E"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    "  </BaseGlyphList>",
                    "  <LayerList>",
                    "    <!-- LayerCount=6 -->",
                    '    <Paint index="0" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="0"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="B"/>',
                    "    </Paint>",
                    '    <Paint index="1" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="1"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="C"/>',
                    "    </Paint>",
                    '    <Paint index="2" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="2"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="D"/>',
                    "    </Paint>",
                    '    <Paint index="3" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="1"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="C"/>',
                    "    </Paint>",
                    '    <Paint index="4" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="2"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="D"/>',
                    "    </Paint>",
                    '    <Paint index="5" Format="10"><!-- PaintGlyph -->',
                    '      <Paint Format="2"><!-- PaintSolid -->',
                    '        <PaletteIndex value="3"/>',
                    '        <Alpha value="1.0"/>',
                    "      </Paint>",
                    '      <Glyph value="F"/>',
                    "    </Paint>",
                    "  </LayerList>",
                    "</COLR>",
                ],
                id="simple-reuse",
            ),
            pytest.param(
                {
                    "A": {
                        "Format": int(ot.PaintFormat.PaintGlyph),
                        "Paint": {
                            "Format": int(ot.PaintFormat.PaintSolid),
                            "PaletteIndex": 0,
                            "Alpha": 1.0,
                        },
                        "Glyph": "B",
                    },
                },
                [
                    "<COLR>",
                    '  <Version value="1"/>',
                    "  <!-- BaseGlyphRecordCount=0 -->",
                    "  <!-- LayerRecordCount=0 -->",
                    "  <BaseGlyphList>",
                    "    <!-- BaseGlyphCount=1 -->",
                    '    <BaseGlyphPaintRecord index="0">',
                    '      <BaseGlyph value="A"/>',
                    '      <Paint Format="10"><!-- PaintGlyph -->',
                    '        <Paint Format="2"><!-- PaintSolid -->',
                    '          <PaletteIndex value="0"/>',
                    '          <Alpha value="1.0"/>',
                    "        </Paint>",
                    '        <Glyph value="B"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    "  </BaseGlyphList>",
                    "</COLR>",
                ],
                [
                    "<COLR>",
                    '  <Version value="1"/>',
                    "  <!-- BaseGlyphRecordCount=0 -->",
                    "  <!-- LayerRecordCount=0 -->",
                    "  <BaseGlyphList>",
                    "    <!-- BaseGlyphCount=1 -->",
                    '    <BaseGlyphPaintRecord index="0">',
                    '      <BaseGlyph value="A"/>',
                    '      <Paint Format="10"><!-- PaintGlyph -->',
                    '        <Paint Format="2"><!-- PaintSolid -->',
                    '          <PaletteIndex value="0"/>',
                    '          <Alpha value="1.0"/>',
                    "        </Paint>",
                    '        <Glyph value="B"/>',
                    "      </Paint>",
                    "    </BaseGlyphPaintRecord>",
                    "  </BaseGlyphList>",
                    "</COLR>",
                ],
                id="no-layer-list",
            ),
        ],
    )
    def test_expandPaintColrLayers(
        self, color_glyphs, ttFont, before_xml, expected_xml
    ):
        colr = buildCOLR(color_glyphs, allowLayerReuse=True)

        assert dump_xml(colr.table, ttFont) == before_xml

        before_layer_count = 0
        reuses_colr_layers = False
        if colr.table.LayerList:
            before_layer_count = len(colr.table.LayerList.Paint)
            reuses_colr_layers = any(
                p.Format == ot.PaintFormat.PaintColrLayers
                for p in colr.table.LayerList.Paint
            )

        COLRVariationMerger.expandPaintColrLayers(colr.table)

        assert dump_xml(colr.table, ttFont) == expected_xml

        after_layer_count = (
            0 if not colr.table.LayerList else len(colr.table.LayerList.Paint)
        )

        if reuses_colr_layers:
            assert not any(
                p.Format == ot.PaintFormat.PaintColrLayers
                for p in colr.table.LayerList.Paint
            )
            assert after_layer_count > before_layer_count
        else:
            assert after_layer_count == before_layer_count

        if colr.table.LayerList:
            assert len({id(p) for p in colr.table.LayerList.Paint}) == after_layer_count


class SparsePositioningMergerTest:
    def test_zero_kern_at_default(self):
        # https://github.com/fonttools/fonttools/issues/3111

        pytest.importorskip("ufo2ft")
        pytest.importorskip("ufoLib2")

        from fontTools.designspaceLib import DesignSpaceDocument
        from ufo2ft import compileVariableTTF
        from ufoLib2 import Font

        ds = DesignSpaceDocument()
        ds.addAxisDescriptor(
            name="wght", tag="wght", minimum=100, maximum=900, default=400
        )
        ds.addSourceDescriptor(font=Font(), location=dict(wght=100))
        ds.addSourceDescriptor(font=Font(), location=dict(wght=400))
        ds.addSourceDescriptor(font=Font(), location=dict(wght=900))

        ds.sources[0].font.newGlyph("a").unicode = ord("a")
        ds.sources[0].font.newGlyph("b").unicode = ord("b")
        ds.sources[0].font.features.text = "feature kern { pos a b b' 100; } kern;"

        ds.sources[1].font.newGlyph("a").unicode = ord("a")
        ds.sources[1].font.newGlyph("b").unicode = ord("b")
        ds.sources[1].font.features.text = "feature kern { pos a b b' 0; } kern;"

        ds.sources[2].font.newGlyph("a").unicode = ord("a")
        ds.sources[2].font.newGlyph("b").unicode = ord("b")
        ds.sources[2].font.features.text = "feature kern { pos a b b' -100; } kern;"

        font = compileVariableTTF(ds, inplace=True)
        b = BytesIO()
        font.save(b)

        assert font["GDEF"].table.VarStore.VarData[0].Item[0] == [100, -100]

    def test_sparse_cursive(self):
        # https://github.com/fonttools/fonttools/issues/3168

        pytest.importorskip("ufo2ft")
        pytest.importorskip("ufoLib2")

        from fontTools.designspaceLib import DesignSpaceDocument
        from ufo2ft import compileVariableTTF
        from ufoLib2 import Font

        ds = DesignSpaceDocument()
        ds.addAxisDescriptor(
            name="wght", tag="wght", minimum=100, maximum=900, default=400
        )
        ds.addSourceDescriptor(font=Font(), location=dict(wght=100))
        ds.addSourceDescriptor(font=Font(), location=dict(wght=400))
        ds.addSourceDescriptor(font=Font(), location=dict(wght=900))

        ds.sources[0].font.newGlyph("a").unicode = ord("a")
        ds.sources[0].font.newGlyph("b").unicode = ord("b")
        ds.sources[0].font.newGlyph("c").unicode = ord("c")
        ds.sources[
            0
        ].font.features.text = """
        feature curs {
          position cursive a <anchor 400 20> <anchor 0 -20>;
          position cursive c <anchor NULL> <anchor 0 -20>;
        } curs;
        """

        ds.sources[1].font.newGlyph("a").unicode = ord("a")
        ds.sources[1].font.newGlyph("b").unicode = ord("b")
        ds.sources[1].font.newGlyph("c").unicode = ord("c")
        ds.sources[
            1
        ].font.features.text = """
        feature curs {
          position cursive a <anchor 500 20> <anchor 0 -20>;
          position cursive b <anchor 50 22> <anchor 0 -10>;
          position cursive c <anchor NULL> <anchor 0 -20>;
        } curs;
        """

        ds.sources[2].font.newGlyph("a").unicode = ord("a")
        ds.sources[2].font.newGlyph("b").unicode = ord("b")
        ds.sources[2].font.newGlyph("c").unicode = ord("c")
        ds.sources[
            2
        ].font.features.text = """
        feature curs {
          position cursive b <anchor 100 40> <anchor 0 -30>;
          position cursive c <anchor NULL> <anchor 0 -20>;
        } curs;
        """

        font = compileVariableTTF(ds, inplace=True)
        b = BytesIO()
        font.save(b)

        assert font["GDEF"].table.VarStore.VarData[0].Item[0] == [-100, 0]



from typing import NamedTuple


class FeatureLibLocation(NamedTuple):
    """A location in a feature file"""

    file: str
    line: int
    column: int

    def __str__(self):
        return f"{self.file}:{self.line}:{self.column}"



"""CFF2 to CFF converter."""

from fontTools.ttLib import TTFont, newTable
from fontTools.misc.cliTools import makeOutputFileName
from fontTools.misc.psCharStrings import T2StackUseExtractor
from fontTools.cffLib import (
    TopDictIndex,
    buildOrder,
    buildDefaults,
    topDictOperators,
    privateDictOperators,
    FDSelect,
)
from .transforms import desubroutinizeCharString
from .specializer import specializeProgram
from .width import optimizeWidths
from collections import defaultdict
import logging


__all__ = ["convertCFF2ToCFF", "main"]


log = logging.getLogger("fontTools.cffLib")


def _convertCFF2ToCFF(cff, otFont):
    """Converts this object from CFF2 format to CFF format. This conversion
    is done 'in-place'. The conversion cannot be reversed.

    The CFF2 font cannot be variable. (TODO Accept those and convert to the
    default instance?)

    This assumes a decompiled CFF2 table. (i.e. that the object has been
    filled via :meth:`decompile` and e.g. not loaded from XML.)"""

    cff.major = 1

    topDictData = TopDictIndex(None)
    for item in cff.topDictIndex:
        # Iterate over, such that all are decompiled
        item.cff2GetGlyphOrder = None
        topDictData.append(item)
    cff.topDictIndex = topDictData
    topDict = topDictData[0]

    if hasattr(topDict, "VarStore"):
        raise ValueError("Variable CFF2 font cannot be converted to CFF format.")

    opOrder = buildOrder(topDictOperators)
    topDict.order = opOrder
    for key in topDict.rawDict.keys():
        if key not in opOrder:
            del topDict.rawDict[key]
            if hasattr(topDict, key):
                delattr(topDict, key)

    charStrings = topDict.CharStrings

    fdArray = topDict.FDArray
    if not hasattr(topDict, "FDSelect"):
        # FDSelect is optional in CFF2, but required in CFF.
        fdSelect = topDict.FDSelect = FDSelect()
        fdSelect.gidArray = [0] * len(charStrings.charStrings)

    defaults = buildDefaults(privateDictOperators)
    order = buildOrder(privateDictOperators)
    for fd in fdArray:
        fd.setCFF2(False)
        privateDict = fd.Private
        privateDict.order = order
        for key in order:
            if key not in privateDict.rawDict and key in defaults:
                privateDict.rawDict[key] = defaults[key]
        for key in privateDict.rawDict.keys():
            if key not in order:
                del privateDict.rawDict[key]
                if hasattr(privateDict, key):
                    delattr(privateDict, key)

    # Add ending operators
    for cs in charStrings.values():
        cs.decompile()
        cs.program.append("endchar")
    for subrSets in [cff.GlobalSubrs] + [
        getattr(fd.Private, "Subrs", []) for fd in fdArray
    ]:
        for cs in subrSets:
            cs.program.append("return")

    # Add (optimal) width to CharStrings that need it.
    widths = defaultdict(list)
    metrics = otFont["hmtx"].metrics
    for glyphName in charStrings.keys():
        cs, fdIndex = charStrings.getItemAndSelector(glyphName)
        if fdIndex == None:
            fdIndex = 0
        widths[fdIndex].append(metrics[glyphName][0])
    for fdIndex, widthList in widths.items():
        bestDefault, bestNominal = optimizeWidths(widthList)
        private = fdArray[fdIndex].Private
        private.defaultWidthX = bestDefault
        private.nominalWidthX = bestNominal
    for glyphName in charStrings.keys():
        cs, fdIndex = charStrings.getItemAndSelector(glyphName)
        if fdIndex == None:
            fdIndex = 0
        private = fdArray[fdIndex].Private
        width = metrics[glyphName][0]
        if width != private.defaultWidthX:
            cs.program.insert(0, width - private.nominalWidthX)

    # Handle stack use since stack-depth is lower in CFF than in CFF2.
    for glyphName in charStrings.keys():
        cs, fdIndex = charStrings.getItemAndSelector(glyphName)
        if fdIndex is None:
            fdIndex = 0
        private = fdArray[fdIndex].Private
        extractor = T2StackUseExtractor(
            getattr(private, "Subrs", []), cff.GlobalSubrs, private=private
        )
        stackUse = extractor.execute(cs)
        if stackUse > 48:  # CFF stack depth is 48
            desubroutinizeCharString(cs)
            cs.program = specializeProgram(cs.program)

    # Unused subroutines are still in CFF2 (ie. lacking 'return' operator)
    # because they were not decompiled when we added the 'return'.
    # Moreover, some used subroutines may have become unused after the
    # stack-use fixup. So we remove all unused subroutines now.
    cff.remove_unused_subroutines()

    mapping = {
        name: ("cid" + str(n).zfill(5) if n else ".notdef")
        for n, name in enumerate(topDict.charset)
    }
    topDict.charset = [
        "cid" + str(n).zfill(5) if n else ".notdef" for n in range(len(topDict.charset))
    ]
    charStrings.charStrings = {
        mapping[name]: v for name, v in charStrings.charStrings.items()
    }

    topDict.ROS = ("Adobe", "Identity", 0)


def convertCFF2ToCFF(font, *, updatePostTable=True):
    if "CFF2" not in font:
        raise ValueError("Input font does not contain a CFF2 table.")
    cff = font["CFF2"].cff
    _convertCFF2ToCFF(cff, font)
    del font["CFF2"]
    table = font["CFF "] = newTable("CFF ")
    table.cff = cff

    if updatePostTable and "post" in font:
        # Only version supported for fonts with CFF table is 0x00030000 not 0x20000
        post = font["post"]
        if post.formatType == 2.0:
            post.formatType = 3.0


def main(args=None):
    """Convert CFF2 OTF font to CFF OTF font"""
    if args is None:
        import sys

        args = sys.argv[1:]

    import argparse

    parser = argparse.ArgumentParser(
        "fonttools cffLib.CFF2ToCFF",
        description="Convert a non-variable CFF2 font to CFF.",
    )
    parser.add_argument(
        "input", metavar="INPUT.ttf", help="Input OTF file with CFF table."
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="OUTPUT.ttf",
        default=None,
        help="Output instance OTF file (default: INPUT-CFF2.ttf).",
    )
    parser.add_argument(
        "--no-recalc-timestamp",
        dest="recalc_timestamp",
        action="store_false",
        help="Don't set the output font's timestamp to the current time.",
    )
    parser.add_argument(
        "--remove-overlaps",
        action="store_true",
        help="Merge overlapping contours and components. Requires skia-pathops",
    )
    parser.add_argument(
        "--ignore-overlap-errors",
        action="store_true",
        help="Don't crash if the remove-overlaps operation fails for some glyphs.",
    )
    loggingGroup = parser.add_mutually_exclusive_group(required=False)
    loggingGroup.add_argument(
        "-v", "--verbose", action="store_true", help="Run more verbosely."
    )
    loggingGroup.add_argument(
        "-q", "--quiet", action="store_true", help="Turn verbosity off."
    )
    options = parser.parse_args(args)

    from fontTools import configLogger

    configLogger(
        level=("DEBUG" if options.verbose else "ERROR" if options.quiet else "INFO")
    )

    import os

    infile = options.input
    if not os.path.isfile(infile):
        parser.error("No such file '{}'".format(infile))

    outfile = (
        makeOutputFileName(infile, overWrite=True, suffix="-CFF")
        if not options.output
        else options.output
    )

    font = TTFont(infile, recalcTimestamp=options.recalc_timestamp, recalcBBoxes=False)

    convertCFF2ToCFF(font)

    if options.remove_overlaps:
        from fontTools.ttLib.removeOverlaps import removeOverlaps
        from io import BytesIO

        log.debug("Removing overlaps")

        stream = BytesIO()
        font.save(stream)
        stream.seek(0)
        font = TTFont(stream, recalcTimestamp=False, recalcBBoxes=False)
        removeOverlaps(
            font,
            ignoreErrors=options.ignore_overlap_errors,
        )

    log.info(
        "Saving %s",
        outfile,
    )
    font.save(outfile)


if __name__ == "__main__":
    import sys

    sys.exit(main(sys.argv[1:]))
